
// The functions for reading a single value from the database using SQLGetData.  There is a different function for
// every data type.

#include "pyodbc.h"
#include "pyodbcmodule.h"
#include "cursor.h"
#include "connection.h"
#include "errors.h"

enum
{
    // The maximum amount of stack we use for unbound variable length columns.  We'll read in this much data to the
    // stack and then copy it to the final buffer when we know the final length.
    MAX_VAR_STACK = 2048,
};

void GetData_init()
{
    PyDateTime_IMPORT;    
}

static PyObject*
GetDataString(Cursor* cur, int iCol)
{
    // Returns a String or Unicode object for character and binary data.

    // NULL terminator notes:
    //
    //  * pinfo->column_size, from SQLDescribeCol, does not include a NULL terminator.  For example, column_size for a
    //    char(10) column would be 10.  (Also, when dealing with wchar_t, it is the number of *characters*, not bytes.)
    //
    //  * When passing a length to PyString_FromStringAndSize and similar Unicode functions, do not add the NULL
    //    terminator -- it will be added automatically.  See objects/stringobject.c
    //
    //  * SQLGetData does not return the NULL terminator in the length indicator.  (Therefore, you can pass this value
    //    directly to the Python string functions.)
    //
    //  * SQLGetData will write a NULL terminator in the output buffer, so you must leave room for it.  You must also
    //    include the NULL terminator in the buffer length passed to SQLGetData.
    //
    // ODBC generalization:
    //  1) Include NULL terminators in input buffer lengths.
    //  2) NULL terminators are not used in data lengths.

    ColumnInfo* pinfo = &cur->colinfos[iCol];

    // Some Unix ODBC drivers do not return the correct length.
    if (pinfo->sql_type == SQL_GUID)
        pinfo->column_size = 36;

    SQLSMALLINT nTargetType;    // ASCII or Unicode?
    int element_size;           // How wide is each character: ASCII/ANSI -> 1, Unicode -> 2 or 4, binary -> 1
    int null_size;              // How much room to add for null terminator: binary -> 0, other -> same as a element_size

    switch (pinfo->sql_type)
    {
    case SQL_CHAR:
    case SQL_VARCHAR:
    case SQL_LONGVARCHAR:
    case SQL_GUID:
        nTargetType  = SQL_C_CHAR;
        element_size = sizeof(char);
        null_size    = sizeof(char);
        break;

    case SQL_WCHAR:
    case SQL_WVARCHAR:
    case SQL_WLONGVARCHAR:
        nTargetType  = SQL_C_WCHAR;
        element_size = Py_UNICODE_SIZE;
        null_size    = Py_UNICODE_SIZE;
        break;
        
    default:
        nTargetType  = SQL_C_BINARY;
        element_size = sizeof(char);
        null_size    = 0;
        break;
    }

    // Since we don't know the value length yet (pinfo->column_size is the maximum length), we'll read ~2K into a
    // buffer on the stack.  In many cases, the entire value will fit and the driver will tell us the exact length (in
    // cbData).

    char abBuffer[MAX_VAR_STACK];
    SQLLEN cbBuffer = _countof(abBuffer);
    SQLLEN cbData   = 0;

    SQLRETURN ret;
    Py_BEGIN_ALLOW_THREADS
    ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nTargetType, abBuffer, cbBuffer, &cbData);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
        return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);

    if (cbData == SQL_NULL_DATA)
        Py_RETURN_NONE;

    // Did we get all of the data?
    // * if SQLSTATE 01004 (data right truncation), we did not.
    // * if the returned length is SQL_NO_TOTAL, we did not.
    // * if the returned length is greater than the buffer we passed in, we did not.

    bool partialData = ret == SQL_SUCCESS_WITH_INFO && HasSqlState(cur->hstmt, "01004");
    partialData |= (cbData == SQL_NO_TOTAL) || cbData > (SQLLEN)(_countof(abBuffer) - null_size);

    if (!partialData)
    {
        if (nTargetType == SQL_C_WCHAR)
            return PyUnicode_FromUnicode((const Py_UNICODE*)abBuffer, cbData / element_size);
        else
            return PyString_FromStringAndSize(abBuffer, cbData);
    }

    //
    // Allocate an object to read into.
    //

    // According to the SQLGetData documentation, the driver is supposed to return the total data length in cbData, not
    // the truncated length (not the amount returned to us).  However, the MS Access driver truncates to 256 and
    // returns 256.
    //
    // We have to know what kind of driver we'll dealing with to be able to do the buffer math, so we'll compare the
    // return value to the buffer size.  If the return value is smaller than the buffer size, we know it is a broken
    // driver like Access (e.g. we give 4000 bytes, and it returns

    SQLLEN cbAllocated;         // does not include NULL
    SQLLEN cbRead;

    if (cbData == SQL_NO_TOTAL)
    {
        // We have no idea, but it is definitely larger than the buffer we've already supplied.
        cbAllocated = cbBuffer * 2;
        cbRead      = cbBuffer - null_size;
    }
    else if (cbData < cbBuffer - null_size)
    {
        // This is some kind of broken driver -- there is room left in the buffer.
        cbAllocated = cbBuffer;
        cbRead      = cbData;
    }
    else
    {
        // We know the total.
        cbAllocated = cbData;
        cbRead      = cbBuffer - null_size;
    }

    char* pch;
    PyObject* result = 0;

    if (nTargetType == SQL_C_WCHAR)
    {
        result = PyUnicode_FromUnicode(0, cbData / element_size);
        pch    = result ? (char*)PyUnicode_AsUnicode(result) : 0;
    }
    else
    {
        result = PyString_FromStringAndSize(0, cbData);
        pch    = result ? PyString_AS_STRING(result) : 0;
    }
        
    if (!result)
        return 0;

    memcpy(pch, abBuffer, cbRead);

    // Drivers are not supposed to include the NULL terminator when telling us the length.
    I(cbRead == 0 || pch[cbRead-element_size] != 0);

    //
    // Loop until SQL_SUCCESS.
    //

    for (;;)
    {
        //
        // Reallocate if necessary.
        //

        // How much room (in bytes) do we have in the current buffer?  We add null_size because both string and
        // unicode objects allocate room for the NULL.
        SQLLEN cbAvailable = cbAllocated - cbRead + null_size;
        
        if (cbAvailable < element_size)
        {
            cbAllocated = (SQLLEN)(cbAllocated * 1.2);

            if (nTargetType == SQL_C_WCHAR)
            {
                if (PyUnicode_Resize(&result, cbAllocated / element_size) == -1)
                    return 0;
                pch = (char*)PyUnicode_AsUnicode(result);
            }
            else
            {
                if (_PyString_Resize(&result, cbAllocated) ==  -1)
                    return 0;
                pch = PyString_AS_STRING(result);
            }
        }

        //
        // Read
        //

        cbData = 0;
        Py_BEGIN_ALLOW_THREADS
        ret = SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nTargetType, pch + cbRead, cbAvailable, &cbData);
        Py_END_ALLOW_THREADS

        if (!SQL_SUCCEEDED(ret) && ret != SQL_NO_DATA)
        {
            Py_DECREF(result);
            return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);
        }

        cbRead += cbData;

        I(cbRead == 0 || pch[cbRead-element_size] != 0);
        I(cbRead <= cbAllocated);

        //
        // If successful, resize and return.
        //

        if (ret == SQL_SUCCESS || ret == SQL_NO_DATA)
        {
            if (cbRead != cbAvailable)
            if (nTargetType == SQL_C_WCHAR)
            {
                if (PyUnicode_Resize(&result, cbRead / element_size) == -1)
                    return 0;
            }
            else
            {
                if (_PyString_Resize(&result, cbRead) ==  -1)
                    return 0;
            }
                
            return result;
        }
    }
}


static PyObject*
GetDataBuffer(Cursor* cur, Py_ssize_t iCol)
{
    PyObject* str = GetDataString(cur, iCol);

    if (str == Py_None)
        return str;

    PyObject* buffer = 0;

    if (str)
    {
        buffer = PyBuffer_FromObject(str, 0, PyString_GET_SIZE(str));
        Py_DECREF(str);         // If no buffer, release it.  If buffer, the buffer owns it.
    }

    return buffer;
}

static PyObject*
GetDataDecimal(Cursor* cur, int iCol)
{
    // The SQL_NUMERIC_STRUCT support is hopeless (SQL Server ignores scale on input parameters and output columns), so
    // we'll rely on the Decimal's string parsing.  Unfortunately, the Decimal author does not pay attention to the
    // locale, so we have to modify the string ourselves.
    //
    // Oracle inserts group separators (commas in US, periods in some countries), so leave room for that too.

    ColumnInfo* pinfo = &cur->colinfos[iCol];

    SQLLEN cbNeeded = pinfo->column_size + 3 +      // sign, decimal, NULL
                      (pinfo->column_size / 3) + 2; // grouping.  I believe this covers all cases.

    SQLLEN cbFetched = 0;
    char* sz = (char*)_alloca(cbNeeded);

    if (sz == 0)
        return PyErr_NoMemory();

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_CHAR, sz, cbNeeded, &cbFetched)))
        return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    // The decimal class requires the decimal to be a period and does not allow thousands separators.  Clean it up.
    //
    // Unfortunately this code only handles single-character values, which might be good enough for decimals and
    // separators, but is certainly not good enough for currency symbols.
    //
    // Note: cbFetched does not include the NULL terminator.

    for (int i = cbFetched - 1; i >=0; i--)
    {
        if (sz[i] == chGroupSeparator || sz[i] == '$' || sz[i] == chCurrencySymbol)
        {
            memmove(&sz[i], &sz[i] + 1, cbFetched - i);
            cbFetched--;
        }
        else if (sz[i] == chDecimal)
        {
            sz[i] = '.';
        }
    }

    return PyObject_CallFunction(decimal_type, "s", sz);
}

static PyObject*
GetDataBit(Cursor* cur, int iCol)
{
    SQLCHAR ch;
    SQLLEN cbFetched;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_BIT, &ch, sizeof(ch), &cbFetched)))
        return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    if (ch == SQL_TRUE)
        Py_RETURN_TRUE;

    Py_RETURN_FALSE;
}

static PyObject*
GetDataLong(Cursor* cur, int iCol)
{
    ColumnInfo* pinfo = &cur->colinfos[iCol];

    SQLINTEGER value = 0;
    SQLLEN cbFetched = 0;

    SQLSMALLINT nCType = pinfo->is_unsigned ? SQL_C_ULONG : SQL_C_LONG;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    if (pinfo->is_unsigned)
        return PyInt_FromLong(*(unsigned long*)&value);

    return PyInt_FromLong(value);
}

static PyObject*
GetDataLongLong(Cursor* cur, int iCol)
{
    ColumnInfo* pinfo = &cur->colinfos[iCol];

    INT64 value = 0;
    SQLLEN cbFetched = 0;

    SQLSMALLINT nCType = pinfo->is_unsigned ? SQL_C_UBIGINT : SQL_C_SBIGINT;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), nCType, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    if (pinfo->is_unsigned)
        return PyLong_FromLongLong(*(UINT64*)&value);

    return PyLong_FromLongLong(*(INT64*)&value);
}

static PyObject*
GetDataDouble(Cursor* cur, int iCol)
{
    double value;
    SQLLEN cbFetched = 0;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_DOUBLE, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    return PyFloat_FromDouble(value);
}


static PyObject*
GetDataTimestamp(Cursor* cur, int iCol)
{
    TIMESTAMP_STRUCT value;

    SQLLEN cbFetched = 0;

    if (!SQL_SUCCEEDED(SQLGetData(cur->hstmt, (SQLSMALLINT)(iCol+1), SQL_C_TYPE_TIMESTAMP, &value, sizeof(value), &cbFetched)))
        return RaiseErrorFromHandle("SQLGetData", cur->cnxn->hdbc, cur->hstmt);

    if (cbFetched == SQL_NULL_DATA)
        Py_RETURN_NONE;

    switch (cur->colinfos[iCol].sql_type)
    {
    case SQL_TYPE_TIME:
        return PyTime_FromTime(value.hour, value.minute, value.second, 0);

    case SQL_TYPE_DATE:
        return PyDate_FromDate(value.year, value.month, value.day);
    }

    // The fraction field is in nanoseconds.
    int micros = value.fraction / 1000;

    return PyDateTime_FromDateAndTime(value.year, value.month, value.day, value.hour, value.minute, value.second, micros);
}


PyObject*
GetData(Cursor* cur, Py_ssize_t iCol)
{
    // Returns an object representing the value in the row/field.  If 0 is returned, an exception has already been set.
    //
    // The data is assumed to be the default C type for the column's SQL type.

    ColumnInfo* pinfo = &cur->colinfos[iCol];

    switch (pinfo->sql_type)
    {
    case SQL_WCHAR:
    case SQL_WVARCHAR:
    case SQL_WLONGVARCHAR:
    case SQL_CHAR:
    case SQL_VARCHAR:
    case SQL_LONGVARCHAR:
    case SQL_GUID:
        return GetDataString(cur, iCol);

    case SQL_BINARY:
    case SQL_VARBINARY:
    case SQL_LONGVARBINARY:
        return GetDataBuffer(cur, iCol);

    case SQL_DECIMAL:
    case SQL_NUMERIC:
    {
        if (decimal_type == 0)
            break;

        return GetDataDecimal(cur, iCol);
    }

    case SQL_BIT:
        return GetDataBit(cur, iCol);

    case SQL_TINYINT:
    case SQL_SMALLINT:
    case SQL_INTEGER:
        return GetDataLong(cur, iCol);

    case SQL_BIGINT:
        return GetDataLongLong(cur, iCol);

    case SQL_REAL:
    case SQL_FLOAT:
    case SQL_DOUBLE:
        return GetDataDouble(cur, iCol);


    case SQL_TYPE_DATE:
    case SQL_TYPE_TIME:
    case SQL_TYPE_TIMESTAMP:
        return GetDataTimestamp(cur, iCol);
    }

    return RaiseErrorV("HY106", ProgrammingError, "ODBC SQL type %d is not yet supported.  column-index=%zd  type=%d",
                       (int)pinfo->sql_type, iCol, (int)pinfo->sql_type);
}
